/* padkey.c
 *
 * BSD PadKey v0.2
 *
 * BSD PadKey is a small kludge of a program for FreeBSD that sends keys
 * to your X display according to which buttons are being pressed on a modified
 * SNES pad connected to the parallel port.
 * 
 * Please view the README document for more information.
 *
 * This program is based on the Linux joy-console.c module. The authors of this 
 * module are listed below:
 *
 *  Copyright (c) 1998 Andree Borrmann
 *  Copyright (c) 1999 John Dahlstrom
 *  Copyright (c) 1999 David Kuder
 *  Copyright (c) 1999 Vojtech Pavlik
 *
 * The following people helped in writing this program:
 *
 *  Copyright (c) 2004 Nicolas Souchu - Wrote functions to access the device 
 *  Copyright (c) 2004 David A. Fulgham - We use a few functions from kbstick.c 
 *  Copyright (c) 2004 Steven D. Lloyd - Assisted me with initial debugging
 *  Copyright (c) 2004 Jason Jennings - Assisted me with button/key handling
 *
 * And I am responsible for this messy hack of a program. My very first useful 
 * program, mind you. Patches that add functionality or clean up code are much 
 * appreciated!
 *
 *  Copyright (c) 2004 Travis Poppe - I kludged this thing together
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * The author of this program, Travis Poppe, may be reached by email 
 * at the following address: tlp@LiquidX.org
 */

#include <sys/param.h>
#include <dev/ppbus/ppi.h>
#include <dev/ppbus/ppbconf.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>

#define JS_PAR_STATUS_INVERT	(0x80)

/* JS_NES_DELAY is set to zero because any other value breaks the program. 
 * Default value was 6 
 */

#define JS_NES_DELAY	0	/* Delay between bits - 6us */

#define JS_SNES_LENGTH	12	/* The SNES true length is 16, but the last 4 bits are unused */

#define JS_SNES_B	0
#define JS_SNES_Y	1
#define JS_SNES_START	2
#define JS_SNES_SELECT	3
#define JS_SNES_UP	4
#define JS_SNES_DOWN	5
#define JS_SNES_LEFT	6
#define JS_SNES_RIGHT	7
#define JS_SNES_A	8
#define JS_SNES_X	9
#define JS_SNES_L	10
#define JS_SNES_R	11

#define JS_NES_POWER	0xfc
#define JS_NES_CLOCK	0x01
#define JS_NES_LATCH	0x02

/* char newLockFileName[256]; */

/* delayForSend is set to zero because any other value causes the keys 
 * not to be sent in real time. Default value was 10
 */

const int delayForSend = 0;    /* Delay in mSec for sending key msgs */

Display *getXDisplay(const char *appName) {
	int	ourEvent, ourError, ourMajor, ourMinor;
	Display *ourDisplay = XOpenDisplay(0);

	if (ourDisplay == NULL) {
		printf("\n ERROR: This program is made to be run in X Windows.\n\n");
/*		perror("Could not open X display"); */
/*		unlink(newLockFileName); */
		exit(1);
	}
	if (!XTestQueryExtension(ourDisplay, &ourEvent, &ourError, &ourMajor, &ourMinor)) {
		printf("%s: ", appName);
		printf("XServer %s doesn't support the Xtest extensions we need...\n", DisplayString(ourDisplay));
		XCloseDisplay(ourDisplay);
		exit(1);
	}
	printf("\nServer %s supports XTest (version %d.%d). Good.\n", DisplayString(ourDisplay), ourMajor, ourMinor);

	XSync(ourDisplay, True);

	return ourDisplay;
}

void	doXKey(Display *ourDisplay, const int keyToSend, int press) {
	if (press)
		XTestFakeKeyEvent(ourDisplay, (unsigned int)keyToSend, True, delayForSend);
	else
		XTestFakeKeyEvent(ourDisplay, (unsigned int)keyToSend, False, delayForSend);
}

static void JS_PAR_DATA_OUT(unsigned char value, int fd) {
	unsigned char val;

	val = value;
	ioctl(fd, PPISDATA, &val);
}

static unsigned char JS_PAR_STATUS(int fd) {
	unsigned char val;

	ioctl(fd, PPIGSTATUS, &val);

	return val;
}
    
static void js_nes_read_packet(int fd, int length, unsigned char *data) {
	int	i;

	JS_PAR_DATA_OUT(JS_NES_POWER | JS_NES_CLOCK | JS_NES_LATCH, fd);
	usleep(JS_NES_DELAY * 2);
	JS_PAR_DATA_OUT(JS_NES_POWER | JS_NES_CLOCK, fd);

	for (i = 0; i < length; i++) {
		usleep(JS_NES_DELAY);
		JS_PAR_DATA_OUT(JS_NES_POWER, fd);
		data[i] = JS_PAR_STATUS(fd) ^ ~JS_PAR_STATUS_INVERT;
		usleep(JS_NES_DELAY);
		JS_PAR_DATA_OUT(JS_NES_POWER | JS_NES_CLOCK, fd);
	}
}

int	main(int argc, char **argv) {
	int	fd, i;
	unsigned char data[JS_SNES_LENGTH];
	int	keyStates[JS_SNES_LENGTH];
	unsigned char XKeys[JS_SNES_LENGTH] = { XK_b, XK_y, XK_p, XK_o, XK_q, XK_w, XK_e, XK_u, XK_a, XK_x, XK_l, XK_r }; 
	Display *ourXDisplay;

	printf("BSD PadKey v0.2 - (c) 2004 Travis Poppe\n\nPlease view the README document for more information.\n"); 

	ourXDisplay = getXDisplay(argv[0]);

	for (i = 0; i < JS_SNES_LENGTH; i++)
		keyStates[i] = 0;

	if ((fd = open("/dev/snespad", O_RDWR, 0)) < 0)
		err(1, "Opening /dev/snespad");


	do {
		int	x, y;

		js_nes_read_packet(fd, JS_SNES_LENGTH, data);

		for (x = 0; x < JS_SNES_LENGTH; x++)
			if (data[x] == 0x40 && keyStates[x] == 0)
				doXKey(ourXDisplay, XKeysymToKeycode(ourXDisplay,XKeys[x]), 1);
			else if (data[x] == 0x0 && keyStates[x] == 1)
				doXKey(ourXDisplay, XKeysymToKeycode(ourXDisplay,XKeys[x]), 0);

		for (y = 0; y < JS_SNES_LENGTH; y++)
			keyStates[y] = (data[y] == 0x40) ? 1 : 0;

		usleep(100);
		XFlush(ourXDisplay);
	} while (1);

	return 0;
}
